==> mars-0.9.2.tar.gz: mars-0.9.2/src/mpu/lib/mars_task_queue.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <string.h>

#include "mars/mars_task_queue.h"
#include "mars/mars_task_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_dma.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_queue queue;

int mars_task_queue_initialize(uint64_t queue_ea,
				uint64_t buffer_ea,
				uint32_t size,
				uint32_t depth,
				uint8_t direction)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(buffer_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET((buffer_ea & MARS_TASK_QUEUE_BUFFER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET((size & MARS_TASK_QUEUE_ENTRY_SIZE_MASK) == 0 &&
			size <= MARS_TASK_QUEUE_ENTRY_SIZE_MAX,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(direction == MARS_TASK_QUEUE_HOST_TO_MPU ||
			direction == MARS_TASK_QUEUE_MPU_TO_HOST ||
			direction == MARS_TASK_QUEUE_MPU_TO_MPU,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize(queue_ea);
	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	queue.mars_context_ea = mars_get_mars_context_ea();
	queue.buffer_ea = buffer_ea;
	queue.push_ea = queue.buffer_ea;
	queue.pop_ea = queue.buffer_ea;
	queue.size = size;
	queue.depth = depth;
	queue.direction = direction;
	queue.count = 0;
	queue.push_wait_count = 0;
	queue.pop_wait_count = 0;

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_count(uint64_t queue_ea, uint32_t *count)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(count, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* return current items in queue */
	*count = queue.count;

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_clear(uint64_t queue_ea)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* signal all waiting tasks that queue is ready for push */
	for (i = 0; i < queue.push_wait_count; i++)
		mars_signal_send(queue.push_wait_id[i]);

	/* flush all ids from push wait list */
	queue.push_wait_count = 0;

	/* clear the queue count and push/pop ea */
	queue.count = 0;
	queue.push_ea = queue.buffer_ea;
	queue.pop_ea = queue.buffer_ea;

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

static void push_update(void)
{
	/* signal waiting task that queue is ready for pop */
	if (queue.pop_wait_count) {
		/* signal waiting task */
		mars_signal_send(queue.pop_wait_id[0]);

		/* flush id from pop wait list */
		queue.pop_wait_count--;
		memmove(&queue.pop_wait_id[0], &queue.pop_wait_id[1],
			sizeof(uint16_t) * queue.pop_wait_count);
	}

	/* increment queue count */
	queue.count++;

	/* increment queue push ea */
	queue.push_ea += queue.size;

	/* wrap to front of queue if necessary */
	if (queue.push_ea == queue.buffer_ea + (queue.size * queue.depth))
		queue.push_ea = queue.buffer_ea;
}

static int push(uint64_t queue_ea, const void *data,
		int try, int begin, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(((uintptr_t)data & MARS_TASK_QUEUE_ENTRY_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		queue.direction == MARS_TASK_QUEUE_MPU_TO_HOST ||
		queue.direction == MARS_TASK_QUEUE_MPU_TO_MPU,
		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue),
		MARS_ERROR_STATE);

	/* queue is full so wait */
	while (queue.count == queue.depth) {
		/* only try so return busy */
		if (try) {
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue);
			return MARS_ERROR_BUSY;
		}

		/* check if push wait list full */
		if (queue.push_wait_count == MARS_TASK_QUEUE_WAIT_MAX) {
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue);
			return MARS_ERROR_LIMIT;
		}

		/* add id to push wait list */
		queue.push_wait_id[queue.push_wait_count]
			= mars_get_workload_id();
		queue.push_wait_count++;

		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

		/* wait for signal */
		mars_signal_wait();

		mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);
	}

	/* begin data transfer into queue */
	mars_dma_put(data, queue.push_ea, queue.size, tag);

	/* only begin data transfer so return */
	if (begin)
		return MARS_SUCCESS;

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data */
	push_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_push(uint64_t queue_ea, const void *data)
{
	return push(queue_ea, data, 0, 0, MARS_DMA_TAG);
}

int mars_task_queue_push_begin(uint64_t queue_ea, const void *data,
				uint32_t tag)
{
	return push(queue_ea, data, 0, 1, tag);
}

int mars_task_queue_push_end(uint64_t queue_ea, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data */
	push_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_try_push(uint64_t queue_ea, const void *data)
{
	return push(queue_ea, data, 1, 0, MARS_DMA_TAG);
}

int mars_task_queue_try_push_begin(uint64_t queue_ea, const void *data,
				uint32_t tag)
{
	return push(queue_ea, data, 1, 1, tag);
}

static void pop_update(void)
{
	/* signal waiting task that queue is ready for push */
	if (queue.push_wait_count) {
		/* signal waiting task */
		mars_signal_send(queue.push_wait_id[0]);

		/* flush id from push wait list */
		queue.push_wait_count--;
		memmove(&queue.push_wait_id[0], &queue.push_wait_id[1],
			sizeof(uint16_t) * queue.push_wait_count);
	}

	/* decrement queue count */
	queue.count--;

	/* increment queue pop ea */
	queue.pop_ea += queue.size;

	/* wrap to front of queue if necessary */
	if (queue.pop_ea == queue.buffer_ea + (queue.size * queue.depth))
		queue.pop_ea = queue.buffer_ea;
}

static int pop(uint64_t queue_ea, void *data,
		int peek, int try, int begin, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(((uintptr_t)data & MARS_TASK_QUEUE_ENTRY_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	struct mars_task_context *task;

	task = (struct mars_task_context *)mars_get_workload_context();

	/* make sure task context has a context save area */
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		queue.direction == MARS_TASK_QUEUE_HOST_TO_MPU ||
		queue.direction == MARS_TASK_QUEUE_MPU_TO_MPU,
		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue),
		MARS_ERROR_STATE);

	/* queue is empty so wait */
	while (!queue.count) {
		/* only try so return busy */
		if (try) {
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue);
			return MARS_ERROR_BUSY;
		}

		/* check if pop wait list full */
		if (queue.pop_wait_count == MARS_TASK_QUEUE_WAIT_MAX) {
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue);
			return MARS_ERROR_LIMIT;
		}

		/* add id to pop wait list */
		queue.pop_wait_id[queue.pop_wait_count]
			= mars_get_workload_id();
		queue.pop_wait_count++;

		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

		/* wait for signal */
		mars_signal_wait();

		mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);
	}

	/* begin data transfer from queue */
	mars_dma_get(data, queue.pop_ea, queue.size, tag);

	/* only begin data transfer so return */
	if (begin)
		return MARS_SUCCESS;

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data only if this is not a peek operation */
	if (!peek)
		pop_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_pop(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 0, 0, 0, MARS_DMA_TAG);
}

int mars_task_queue_pop_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 0, 0, 1, tag);
}

int mars_task_queue_pop_end(uint64_t queue_ea, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data */
	pop_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_try_pop(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 0, 1, 0, MARS_DMA_TAG);
}

int mars_task_queue_try_pop_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 0, 1, 1, tag);
}

int mars_task_queue_peek(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 1, 0, 0, MARS_DMA_TAG);
}

int mars_task_queue_peek_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 1, 0, 1, tag);
}

int mars_task_queue_peek_end(uint64_t queue_ea, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	/* wait for dma completion */
	mars_dma_wait(tag);

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_try_peek(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 1, 1, 0, MARS_DMA_TAG);
}

int mars_task_queue_try_peek_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 1, 1, 1, tag);
}

==> mars-08.06.26.tar.gz: mars-08.06.26/src/mpu/lib/mars_task_queue.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <string.h>

#include "mars/mars_task_queue.h"
#include "mars/mars_workload_types.h"
#include "mars/mars_syscalls.h"
#include "mars/mars_dma.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

static struct mars_task_queue queue;

int mars_task_queue_initialize(uint64_t queue_ea,
				uint64_t buffer_ea,
				uint32_t size,
				uint32_t depth,
				uint8_t direction)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(buffer_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET((buffer_ea & MARS_TASK_QUEUE_BUFFER_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);
	MARS_CHECK_RET((size & MARS_TASK_QUEUE_ENTRY_SIZE_MASK) == 0 &&
			size <= MARS_TASK_QUEUE_ENTRY_SIZE_MAX,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(direction == MARS_TASK_QUEUE_HOST_TO_MPU ||
			direction == MARS_TASK_QUEUE_MPU_TO_HOST ||
			direction == MARS_TASK_QUEUE_MPU_TO_MPU,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize(queue_ea);
	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	queue.mars_context_ea = mars_get_mars_context_ea();
	queue.buffer_ea = buffer_ea;
	queue.push_ea = queue.buffer_ea;
	queue.pop_ea = queue.buffer_ea;
	queue.size = size;
	queue.depth = depth;
	queue.direction = direction;
	queue.count = 0;
	queue.push_wait_count = 0;
	queue.pop_wait_count = 0;

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_count(uint64_t queue_ea, uint32_t *count)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(count, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* return current items in queue */
	*count = queue.count;

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_clear(uint64_t queue_ea)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK) == 0,
			MARS_ERROR_ALIGN);

	int i;

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* clear the queue count and push/pop ea */
	queue.count = 0;
	queue.push_ea = queue.buffer_ea;
	queue.pop_ea = queue.buffer_ea;

	/* signal all waiting tasks that queue is ready for push */
	for (i = 0; i < queue.push_wait_count; i++)
		mars_signal_send(queue.push_wait_id[i]);

	/* flush all ids from push wait list */
	queue.push_wait_count = 0;

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

static void push_update(void)
{
	/* increment queue count */
	queue.count++;

	/* increment queue push ea */
	queue.push_ea += queue.size;

	/* wrap to front of queue if necessary */
	if (queue.push_ea == queue.buffer_ea + (queue.size * queue.depth))
		queue.push_ea = queue.buffer_ea;

	/* signal waiting task that queue is ready for pop */
	if (queue.pop_wait_count) {
		/* signal waiting task */
		mars_signal_send(queue.pop_wait_id[0]);

		/* flush id from pop wait list */
		queue.pop_wait_count--;
		memmove(&queue.pop_wait_id[0], &queue.pop_wait_id[1],
			sizeof(uint16_t) * queue.pop_wait_count);
	}
}

static int push(uint64_t queue_ea, const void *data,
		int try, int begin, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(((uintptr_t)data & MARS_TASK_QUEUE_ENTRY_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		queue.direction == MARS_TASK_QUEUE_MPU_TO_HOST ||
		queue.direction == MARS_TASK_QUEUE_MPU_TO_MPU,
		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue),
		MARS_ERROR_STATE);

	/* queue is full so wait */
	while (queue.count == queue.depth) {
		/* only try so return busy */
		if (try) {
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue);
			return MARS_ERROR_BUSY;
		}

		/* check if push wait list full */
		MARS_CHECK_CLEANUP_RET(
			queue.push_wait_count < MARS_TASK_QUEUE_WAIT_MAX,
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue),
			MARS_ERROR_LIMIT);

		/* add id to push wait list */
		queue.push_wait_id[queue.push_wait_count]
			= mars_get_workload_id();
		queue.push_wait_count++;

		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

		/* wait for signal */
		mars_signal_wait();

		mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);
	}

	/* begin data transfer into queue */
	mars_dma_put(data, queue.push_ea, queue.size, tag);

	/* only begin data transfer so return */
	if (begin)
		return MARS_SUCCESS;

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data */
	push_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_push(uint64_t queue_ea, const void *data)
{
	return push(queue_ea, data, 0, 0, MARS_DMA_TAG);
}

int mars_task_queue_push_begin(uint64_t queue_ea, const void *data,
				uint32_t tag)
{
	return push(queue_ea, data, 0, 1, tag);
}

int mars_task_queue_push_end(uint64_t queue_ea, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data */
	push_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_try_push(uint64_t queue_ea, const void *data)
{
	return push(queue_ea, data, 1, 0, MARS_DMA_TAG);
}

int mars_task_queue_try_push_begin(uint64_t queue_ea, const void *data,
				uint32_t tag)
{
	return push(queue_ea, data, 1, 1, tag);
}

static void pop_update(void)
{
	/* decrement queue count */
	queue.count--;

	/* increment queue pop ea */
	queue.pop_ea += queue.size;

	/* wrap to front of queue if necessary */
	if (queue.pop_ea == queue.buffer_ea + (queue.size * queue.depth))
		queue.pop_ea = queue.buffer_ea;

	/* signal waiting task that queue is ready for push */
	if (queue.push_wait_count) {
		/* signal waiting task */
		mars_signal_send(queue.push_wait_id[0]);

		/* flush id from push wait list */
		queue.push_wait_count--;
		memmove(&queue.push_wait_id[0], &queue.push_wait_id[1],
			sizeof(uint16_t) * queue.push_wait_count);
	}
}

static int pop(uint64_t queue_ea, void *data,
		int peek, int try, int begin, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(((uintptr_t)data & MARS_TASK_QUEUE_ENTRY_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	struct mars_task_context *task;

	/* make sure task context has a context save area */
	task = &mars_get_workload_context()->type_context.task;
	MARS_CHECK_RET(task->context_save_size && task->context_save_area,
			MARS_ERROR_FORMAT);

	mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);

	/* check for valid direction */
	MARS_CHECK_CLEANUP_RET(
		queue.direction == MARS_TASK_QUEUE_HOST_TO_MPU ||
		queue.direction == MARS_TASK_QUEUE_MPU_TO_MPU,
		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue),
		MARS_ERROR_STATE);

	/* queue is empty so wait */
	while (!queue.count) {
		/* only try so return busy */
		if (try) {
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue);
			return MARS_ERROR_BUSY;
		}

		/* check if pop wait list full */
		MARS_CHECK_CLEANUP_RET(
			queue.pop_wait_count < MARS_TASK_QUEUE_WAIT_MAX,
			mars_mutex_unlock_put(queue_ea,
				(struct mars_mutex *)&queue),
			MARS_ERROR_LIMIT);

		/* add id to pop wait list */
		queue.pop_wait_id[queue.pop_wait_count]
			= mars_get_workload_id();
		queue.pop_wait_count++;

		mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

		/* wait for signal */
		mars_signal_wait();

		mars_mutex_lock_get(queue_ea, (struct mars_mutex *)&queue);
	}

	/* begin data transfer from queue */
	mars_dma_get(data, queue.pop_ea, queue.size, tag);

	/* only begin data transfer so return */
	if (begin)
		return MARS_SUCCESS;

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data only if this is not a peek operation */
	if (!peek)
		pop_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_pop(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 0, 0, 0, MARS_DMA_TAG);
}

int mars_task_queue_pop_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 0, 0, 1, tag);
}

int mars_task_queue_pop_end(uint64_t queue_ea, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	/* wait for dma completion */
	mars_dma_wait(tag);

	/* update queue data */
	pop_update();

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_try_pop(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 0, 1, 0, MARS_DMA_TAG);
}

int mars_task_queue_try_pop_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 0, 1, 1, tag);
}

int mars_task_queue_peek(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 1, 0, 0, MARS_DMA_TAG);
}

int mars_task_queue_peek_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 1, 0, 1, tag);
}

int mars_task_queue_peek_end(uint64_t queue_ea, uint32_t tag)
{
	MARS_CHECK_RET(queue_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET((queue_ea & MARS_TASK_QUEUE_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(tag <= MARS_DMA_TAG_MAX, MARS_ERROR_PARAMS);

	/* wait for dma completion */
	mars_dma_wait(tag);

	mars_mutex_unlock_put(queue_ea, (struct mars_mutex *)&queue);

	return MARS_SUCCESS;
}

int mars_task_queue_try_peek(uint64_t queue_ea, void *data)
{
	return pop(queue_ea, data, 1, 1, 0, MARS_DMA_TAG);
}

int mars_task_queue_try_peek_begin(uint64_t queue_ea, void *data, uint32_t tag)
{
	return pop(queue_ea, data, 1, 1, 1, tag);
}
